﻿using Kivii.Linq;
using Kivii.Linq.Legacy;
using Kivii.Samples.Entities;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Kivii.Samples
{
    public static class SampleExtension
    {
        /// <summary>
        /// 主样品创建,传入要创建的样品信息和地点配置信息
        /// </summary>
        /// <param name="sample"></param>
        /// <param name="location"></param>
        /// <param name="route"></param>
        /// <returns></returns>
        public static Sample Accept(this Sample sample, Location location, Route route = null)
        {
            (sample == null).ThrowIfTrue("缺少要创建的样品实例!");
            (sample.Name.IsNullOrEmpty()).ThrowIfTrue("缺少样品名称信息!");
            (location == null).ThrowIfTrue("缺少目的地点配置信息!");
            var conn = KiviiContext.GetOpenedDbConnection<Sample>();
            conn.InitEntityType<Location>();
            conn.InitEntityType<Route>();
            if (location.Kvid == Guid.Empty)
            {
                (location.InternalCode.IsNullOrEmpty()).ThrowIfTrue("缺少目的地点配置InternalCode信息");
                location = conn.Single<Location>(o => o.InternalCode == location.InternalCode);
                (location == null).ThrowIfTrue($"未找到{location.InternalCode}目的地点配置信息!");
            }

            //var local = conn.Single<Location>(o => o.InternalCode == location.InternalCode);
            //(local == null).ThrowIfTrue($"未找到{location.InternalCode}目的地点配置信息!");
            #region 数据验证
            if (!sample.BizId.IsNullOrEmpty())
            {
                (conn.Exists<Sample>(o => o.BizId == sample.BizId && o.RootKvid == o.Kvid && o.Name == sample.Name)).ThrowIfTrue("已存在的主样品!");
            }
            else if (sample.BizKvid != Guid.Empty)
            {
                (conn.Exists<Sample>(o => o.BizKvid == sample.BizKvid && o.RootKvid == o.Kvid && o.Name == sample.Name)).ThrowIfTrue("已存在的主样品!");
            }
            #endregion
            var trans = conn.OpenTransaction();
            try
            {
                #region 创建样品信息
                var sampleKvid = Guid.NewGuid();
                if (sample.OwnerKvid == Guid.Empty)
                {
                    sample.OwnerKvid = KiviiContext.CurrentMember.DepartmentKvid;
                    sample.OwnerName = KiviiContext.CurrentMember.DepartmentName;
                }
                sample.RootKvid = sampleKvid;
                sample.Kvid = sampleKvid;
                sample.ParentKvid = Guid.Empty;
                sample.PackageKvid = Guid.Empty;
                sample.Type = SampleType.Master;
                sample.OperateTime = sample.OperateTime == DateTime.MinValue ? DateTime.Now : sample.OperateTime;
                conn.Insert(sample);
                #endregion

                #region 生成样品路由信息
                if (route == null) route = new Route();
                route.SampleRootKvid = sample.RootKvid;
                route.SampleKvid = sample.Kvid;
                route.SampleParentKvid = sample.ParentKvid;
                route.SampleName = sample.Name;
                route.BizId = sample.BizId;
                route.BizKvid = sample.BizKvid;
                route.BizType = sample.BizType;
                route.Title = route.Title.IsNullOrEmpty() ? $"样品[{sample.Name}]录入" : route.Title;
                route.Type = route.Type.IsNullOrEmpty() ? sample.SourceType : route.Type;
                route.CurrentLocationKvid = location.Kvid;
                route.CurrentLocationTitle = location.Title;
                route.OperateTime = route.OperateTime == DateTime.MinValue ? DateTime.Now : route.OperateTime;
                if (route.OperatorName.IsNullOrEmpty())
                {
                    route.OperatorName = KiviiContext.CurrentMember.FullName;
                    route.OperatorKvid = KiviiContext.CurrentMember.Kvid;
                }
                conn.Insert(route);
                #endregion
                trans.Commit();
            }
            catch (Exception ex)
            {
                trans.Rollback();
                throw ex;
            }
            return sample;
        }

        /// <summary>
        /// 样品流转,此方法用于样品从一个地点到另一个地点,只生成route数据以及回写上级route的Next信息
        /// </summary>
        /// <param name="sample"></param>
        /// <param name="location"></param>
        /// <param name="route">下一地点的路由信息组织好传入,用于创建</param>
        /// <returns></returns>
        public static Sample Assign(this Sample sample, Location location, Route route = null, IDbConnection conn = null)
        {
            (sample == null).ThrowIfTrue("请传入需要流转的样品信息!");
            (location == null).ThrowIfTrue("缺少目的地点配置信息!");
            bool useTransaction = conn == null;//是否启用内部事务
            if (conn == null) conn = KiviiContext.GetOpenedDbConnection<Sample>();
            //var conn = KiviiContext.GetOpenedDbConnection<Sample>();
            conn.InitEntityType<Location>();
            if (location.Kvid == Guid.Empty)
            {
                (location.InternalCode.IsNullOrEmpty()).ThrowIfTrue("缺少目的地点配置InternalCode信息");
                location = conn.Single<Location>(o => o.InternalCode == location.InternalCode);
                (location == null).ThrowIfTrue($"未找到{location.InternalCode}目的地点配置信息!");
            }
            var existSample = conn.SingleById<Sample>(sample.Kvid);
            existSample.ThrowIfNull("未找到此样品信息!");
            //找出当前样品所在路由地点,根据Next为空说明就在此地
            var currentRoute = existSample.GetCurrentRoute();//conn.Single<Route>(o => o.SampleKvid == existSample.Kvid && o.NextLocationKvid == Guid.Empty);

            IDbTransaction trans = null;//事务，如果外部来了Connection，不生成事务
            if (useTransaction)
            {
                trans = conn.OpenTransaction();
            }
            //var trans = conn.OpenTransaction();
            try
            {
                #region 生成新的路由信息,更新上一个路由信息
                Route newRoute = new Route();
                if (route != null) newRoute.PopulateWith(route);
                newRoute.SampleRootKvid = existSample.RootKvid;
                newRoute.SampleKvid = existSample.Kvid;
                newRoute.SampleParentKvid = existSample.ParentKvid;
                newRoute.SampleName = existSample.Name;
                newRoute.BizId = existSample.BizId;
                newRoute.BizKvid = existSample.BizKvid;
                newRoute.BizType = existSample.BizType;
                newRoute.Title = newRoute.Title.IsNullOrEmpty() ? $"样品[{existSample.Name}]流转" : newRoute.Title;
                newRoute.Type = newRoute.Type.IsNullOrEmpty() ? "流转" : newRoute.Type;
                newRoute.CurrentLocationKvid = location.Kvid;
                newRoute.CurrentLocationTitle = location.Title;
                newRoute.OperateTime = newRoute.OperateTime == DateTime.MinValue ? DateTime.Now : newRoute.OperateTime;
                existSample.AssignTime = newRoute.OperateTime;
                if (newRoute.OperatorName.IsNullOrEmpty())
                {
                    newRoute.OperatorName = KiviiContext.CurrentMember.FullName;
                    newRoute.OperatorKvid = KiviiContext.CurrentMember.Kvid;
                }
                conn.Insert(newRoute);

                if (currentRoute != null)
                {
                    currentRoute.NextLocationKvid = location.Kvid;
                    currentRoute.NextLocationTitle = location.Title;
                    currentRoute.AddOnlyProperties(o => o.NextLocationKvid);
                    currentRoute.AddOnlyProperties(o => o.NextLocationTitle);
                    conn.UpdateOnly(currentRoute);
                }
                #endregion
                trans?.Commit();
            }
            catch (Exception ex)
            {
                trans?.Rollback();
                throw ex;
            }
            return existSample;
        }

        /// <summary>
        /// 样品拆分
        /// </summary>
        /// <param name="sample"></param>
        /// <param name="splitSamples"></param>
        /// <param name="location"></param>
        /// <returns></returns>
        public static List<Sample> Split(this Sample sample, List<Sample> splitSamples, Location location)
        {
            (sample == null).ThrowIfTrue("请传入需要拆分的主样品信息!");
            splitSamples.ThrowIfNullOrEmpty("缺少拆分的样品信息");
            (location == null).ThrowIfTrue("缺少目的地点配置信息!");
            var conn = KiviiContext.GetOpenedDbConnection<Sample>();
            conn.InitEntityType<Location>();
            if (location.Kvid == Guid.Empty)
            {
                (location.InternalCode.IsNullOrEmpty()).ThrowIfTrue("缺少目的地点配置InternalCode信息");
                location = conn.Single<Location>(o => o.InternalCode == location.InternalCode);
                (location == null).ThrowIfTrue($"未找到{location.InternalCode}目的地点配置信息!");
            }
            var existSample = conn.SingleById<Sample>(sample.Kvid);
            existSample.ThrowIfNull("未找到此样品信息!");
            (existSample.Type != SampleType.Master && existSample.Type != SampleType.Node).ThrowIfTrue("不支持拆分的样品类型");
            //先获取当前的最后一个路由地址
            var currentRoute = existSample.GetCurrentRoute();
            if (currentRoute != null)
            {
                //判断当前拆分地点和此样品所在地一致  要是不一致要先流转到此地
                if (currentRoute.CurrentLocationKvid != location.Kvid)
                {
                    existSample = existSample.Assign(location, null);
                }
            }
            var rtns = new List<Sample>();

            var trans = conn.OpenTransaction();
            try
            {
                
                foreach (var item in splitSamples)
                {
                    var sampleKvid = Guid.NewGuid();
                    if (item.OwnerKvid == Guid.Empty)
                    {
                        item.OwnerKvid = KiviiContext.CurrentMember.DepartmentKvid;
                        item.OwnerName = KiviiContext.CurrentMember.DepartmentName;
                    }
                    item.Kvid = sampleKvid;
                    item.RootKvid = existSample.RootKvid;
                    item.ParentKvid = existSample.Kvid;
                    item.PackageKvid = Guid.Empty;
                    item.Type = SampleType.Node;
                    item.OperateTime = item.OperateTime == DateTime.MinValue ? DateTime.Now : item.OperateTime;
                    conn.Insert(item);
                    rtns.Add(item);

                    var route = new Route();
                    route.SampleRootKvid = item.RootKvid;
                    route.SampleKvid = item.Kvid;
                    route.SampleParentKvid = item.ParentKvid;
                    route.SampleName = item.Name;
                    route.BizId = item.BizId;
                    route.BizKvid = item.BizKvid;
                    route.BizType = item.BizType;
                    route.Title = $"样品[{existSample.Name}]拆分";
                    route.Type = "拆分";
                    route.CurrentLocationKvid = location.Kvid;
                    route.CurrentLocationTitle = location.Title;
                    route.OperateTime = DateTime.Now;
                    if (route.OperatorName.IsNullOrEmpty())
                    {
                        route.OperatorName = KiviiContext.CurrentMember.FullName;
                        route.OperatorKvid = KiviiContext.CurrentMember.Kvid;
                    }
                    conn.Insert(route);
                }
                trans.Commit();
            }
            catch (Exception ex)
            {
                trans.Rollback();
                throw ex;
            }
            return rtns;
        }

        /// <summary>
        /// 包裹清除
        /// </summary>
        /// <param name="packages"></param>
        /// <param name="location"></param>
        /// <returns></returns>
        public static List<Sample> PackageClear(this List<Sample> packages, Location location, Route route = null, IDbConnection conn = null)
        {
            packages.ThrowIfNullOrEmpty("请传入要清理的包裹!");
            (location == null).ThrowIfTrue("缺少目的地点配置信息!");
            bool useTransaction = conn == null;//是否启用内部事务
            if (conn == null) conn = KiviiContext.GetOpenedDbConnection<Sample>();
            conn.InitEntityType<Location>();
            if (location.Kvid == Guid.Empty)
            {
                (location.InternalCode.IsNullOrEmpty()).ThrowIfTrue("缺少目的地点配置InternalCode信息");
                location = conn.Single<Location>(o => o.InternalCode == location.InternalCode);
                (location == null).ThrowIfTrue($"未找到{location.InternalCode}目的地点配置信息!");
            }
            var exists = conn.SelectByIds<Sample>(packages.ConvertAll(p => p.Kvid));
            exists.ThrowIfNullOrEmpty("未找到要清理的包裹!");
            exists.Exists(o => o.Type != SampleType.Package).ThrowIfTrue("存在不是包裹的信息!");
            var samples = conn.Select<Sample>(o => Sql.In(o.PackageKvid, exists.ConvertAll(p => p.Kvid)));

            var rtns = new List<Sample>();
            IDbTransaction trans = null;//事务，如果外部来了Connection，不生成事务
            if (useTransaction)
            {
                trans = conn.OpenTransaction();
            }
            try
            {
                var updatePackagess = conn.From<Sample>();
                updatePackagess = updatePackagess.Update(o => new { o.PackageKvid, o.PackageName,o.Status });
                updatePackagess = updatePackagess.Where(o => Sql.In(o.Kvid, exists.ConvertAll(p => p.Kvid)));
                conn.UpdateOnly<Sample>(new Sample { PackageKvid = Guid.Empty, PackageName = string.Empty, Status = (int)PackageStatus.Empty }, updatePackagess);
                if (!samples.IsNullOrEmpty())
                {
                    var updateSampless = conn.From<Sample>();
                    updateSampless = updateSampless.Update(o => new { o.PackageKvid, o.PackageName });
                    updateSampless = updateSampless.Where(o => Sql.In(o.Kvid, samples.ConvertAll(p => p.Kvid)));
                    conn.UpdateOnly<Sample>(new Sample { PackageKvid = Guid.Empty, PackageName = string.Empty }, updateSampless);
                    foreach (var item in samples)
                    {
                        var package = exists.FirstOrDefault(o => o.Kvid == item.PackageKvid);
                        var currentRoute = item.GetCurrentRoute();//conn.Single<Route>(o => o.SampleKvid == existSample.Kvid && o.NextLocationKvid == Guid.Empty);
                        if (route == null) route = new Route();
                        route.SampleRootKvid = item.RootKvid;
                        route.SampleKvid = item.Kvid;
                        route.SampleParentKvid = item.ParentKvid;
                        route.SampleName = item.Name;
                        route.BizId = item.BizId;
                        route.BizKvid = item.BizKvid;
                        route.BizType = item.BizType;
                        route.Title = route.Title.IsNullOrEmpty() ? $"出库报废,样品[{item.Name}]已报废" : route.Title;
                        route.Type = route.Type.IsNullOrEmpty() ? "出库报废" : route.Type;
                        if (package != null) route.Summary = $"原货区:{package.PackageName},箱号:{package.Name}";
                        route.CurrentLocationKvid = location.Kvid;
                        route.CurrentLocationTitle = location.Title;
                        route.OperateTime = route.OperateTime == DateTime.MinValue ? DateTime.Now : route.OperateTime;
                        if (route.OperatorName.IsNullOrEmpty())
                        {
                            route.OperatorName = KiviiContext.CurrentMember.FullName;
                            route.OperatorKvid = KiviiContext.CurrentMember.Kvid;
                        }
                        conn.Insert(route);

                        if (currentRoute != null)
                        {
                            currentRoute.NextLocationKvid = location.Kvid;
                            currentRoute.NextLocationTitle = location.Title;
                            currentRoute.AddOnlyProperties(o => o.NextLocationKvid);
                            currentRoute.AddOnlyProperties(o => o.NextLocationTitle);
                            conn.UpdateOnly(currentRoute);
                        }
                    }
                    rtns.AddRange(samples);
                }
                trans?.Commit();
            }
            catch (Exception ex)
            {
                trans?.Rollback();
                throw ex;
            }
            return rtns;
        }

        /// <summary>
        /// 样品打包
        /// </summary>
        /// <param name="samples"></param>
        /// <param name="package">生成的包裹信息</param>
        /// <param name="location"></param>
        /// <param name="packageRoute"></param>
        /// <returns></returns>
        public static List<Sample> Packaging(this List<Sample> samples, Sample package, Location location, Route packageRoute = null, IDbConnection conn = null)
        {
            samples.ThrowIfNullOrEmpty("请传入要打包的样品信息!");
            package.ThrowIfNull("请传入包裹信息!");
            (location == null).ThrowIfTrue("缺少目的地点配置信息!");
            bool useTransaction = conn == null;//是否启用内部事务
            if (conn == null) conn = KiviiContext.GetOpenedDbConnection<Sample>();
            conn.InitEntityType<Location>();
            if (location.Kvid == Guid.Empty)
            {
                (location.InternalCode.IsNullOrEmpty()).ThrowIfTrue("缺少目的地点配置InternalCode信息");
                location = conn.Single<Location>(o => o.InternalCode == location.InternalCode);
                (location == null).ThrowIfTrue($"未找到{location.InternalCode}目的地点配置信息!");
            }
            var exists = conn.SelectByIds<Sample>(samples.ConvertAll(p => p.Kvid));
            exists.ThrowIfNullOrEmpty("未找到要打包的样品信息!");
            exists.Exists(o => o.PackageKvid != Guid.Empty).ThrowIfTrue("存在样品已经打包!");
            var packageArea = package.PackageName;
            package.PackageName = string.Empty;
            //var local = conn.Single<Location>(o => o.InternalCode == location.InternalCode);
            //(local == null).ThrowIfTrue($"未找到{location.InternalCode}目的地点配置信息!");
            //var existPackage = conn.SingleById<Sample>(package.Kvid);
            //if (existPackage != null)
            //{
            //    if (existPackage.Type == SampleType.Package) package = existPackage;
            //}
            //var currentRoutes = conn.Select<Route>(o => Sql.In(o.SampleKvid, exists.ConvertAll(p => p.Kvid)) && o.NextLocationKvid == Guid.Empty);
            var rtns = new List<Sample>();
            IDbTransaction trans = null;//事务，如果外部来了Connection，不生成事务
            if (useTransaction)
            {
                trans = conn.OpenTransaction();
            }
            //var trans = conn.OpenTransaction();
            try
            {
                #region 生成包裹和路由
                //要是包裹不存在就要创建一个新的包裹,要是存在就直接用已有的包裹
                if (package.Kvid == Guid.Empty)
                {
                    var packageKvid = Guid.NewGuid();
                    if (package.OwnerKvid == Guid.Empty)
                    {
                        package.OwnerKvid = KiviiContext.CurrentMember.DepartmentKvid;
                        package.OwnerName = KiviiContext.CurrentMember.DepartmentName;
                    }
                    package.RootKvid = packageKvid;
                    package.Kvid = packageKvid;
                    package.ParentKvid = Guid.Empty;
                    package.PackageKvid = Guid.Empty;
                    package.Type = SampleType.Package;
                    package.OperateTime = package.OperateTime == DateTime.MinValue ? DateTime.Now : package.OperateTime;
                    package.Status = (int)PackageStatus.Full;
                    conn.Insert(package);
                    Route newRoute = new Route();
                    if (packageRoute != null) newRoute.PopulateWith(packageRoute);
                    //if (packageRoute == null) packageRoute = new Route();
                    newRoute.SampleRootKvid = package.RootKvid;
                    newRoute.SampleKvid = package.Kvid;
                    newRoute.SampleParentKvid = package.ParentKvid;
                    newRoute.SampleName = package.Name;
                    newRoute.BizId = package.BizId;
                    newRoute.BizKvid = package.BizKvid;
                    newRoute.BizType = package.BizType;
                    newRoute.Title = newRoute.Title.IsNullOrEmpty() ? $"样品打包至:{package.Name}{(packageArea.IsNullOrEmpty() ? "" : $"({packageArea})")}" : newRoute.Title;
                    newRoute.Type = newRoute.Type.IsNullOrEmpty() ? "打包" : newRoute.Type;
                    newRoute.CurrentLocationKvid = location.Kvid;
                    newRoute.CurrentLocationTitle = location.Title;
                    newRoute.OperateTime = newRoute.OperateTime == DateTime.MinValue ? DateTime.Now : newRoute.OperateTime;
                    if (newRoute.OperatorName.IsNullOrEmpty())
                    {
                        newRoute.OperatorName = KiviiContext.CurrentMember.FullName;
                        newRoute.OperatorKvid = KiviiContext.CurrentMember.Kvid;
                    }
                    conn.Insert(newRoute);
                }
                else
                {
                    package.Status = (int)PackageStatus.Full;
                    package.AddOnlyProperties(o => o.Status);
                    conn.UpdateOnly(package);
                }
                #endregion

                foreach (var item in exists)
                {
                    var currentRoute = item.GetCurrentRoute();
                    //if (currentRoute != null)
                    //{
                    //    //判断当前拆分地点和此样品所在地一致  要是不一致要先流转到此地
                    //    if (currentRoute.CurrentLocationKvid != location.Kvid)
                    //    {
                    //        item.Assign(location, null, conn);
                    //        currentRoute = item.GetCurrentRoute();
                    //    }
                    //}

                    //更新样品的包裹Kvid
                    item.PackageName = $"{package.Name}{(packageArea.IsNullOrEmpty() ? "" : $"({packageArea})")}";
                    item.AddOnlyProperties(o => o.PackageName);
                    item.PackageKvid = package.Kvid;
                    item.AddOnlyProperties(o => o.PackageKvid);
                    conn.UpdateOnly(item);
                    item.RemoveAllOnlyProperties();
                    rtns.Add(item);
                    //如果当前样品存在当前路由信息就更新
                    //if (!currentRoutes.IsNullOrEmpty())
                    //{
                    //    var currentRoute = currentRoutes.FirstOrDefault(o => o.SampleKvid == item.Kvid);
                    if (currentRoute != null)
                    {
                        currentRoute.NextLocationKvid = location.Kvid;
                        currentRoute.NextLocationTitle = location.Title;
                        currentRoute.AddOnlyProperties(o => o.NextLocationKvid);
                        currentRoute.AddOnlyProperties(o => o.NextLocationTitle);
                        conn.UpdateOnly(currentRoute);
                    }
                    //}

                    //生成新的路由信息
                    var route = new Route();
                    route.SampleRootKvid = item.RootKvid;
                    route.SampleKvid = item.Kvid;
                    route.SampleParentKvid = item.ParentKvid;
                    route.SampleName = item.Name;
                    route.BizId = item.BizId;
                    route.BizKvid = item.BizKvid;
                    route.BizType = item.BizType;
                    route.Title = $"样品[{item.Name}]打包[{package.Name}]";
                    route.Type = "打包";
                    route.CurrentLocationKvid = location.Kvid;
                    route.CurrentLocationTitle = location.Title;
                    route.OperateTime = DateTime.Now;
                    if (route.OperatorName.IsNullOrEmpty())
                    {
                        route.OperatorName = KiviiContext.CurrentMember.FullName;
                        route.OperatorKvid = KiviiContext.CurrentMember.Kvid;
                    }
                    conn.Insert(route);
                }


                trans?.Commit();
            }
            catch (Exception ex)
            {
                trans?.Rollback();
                throw ex;
            }
            return rtns;
        }

        /// <summary>
        /// 包裹拆分
        /// </summary>
        /// <param name="sample"></param>
        /// <param name="location"></param>
        /// <param name="route"></param>
        /// <returns></returns>
        public static Sample UnPackaging(this Sample sample, Location location, Route route = null, IDbConnection conn = null)
        {
            sample.ThrowIfNull("请传入要取出的样品信息!");
            (location == null).ThrowIfTrue("缺少目的地点配置信息!");
            bool useTransaction = conn == null;//是否启用内部事务
            if (conn == null) conn = KiviiContext.GetOpenedDbConnection<Sample>();
            conn.InitEntityType<Location>();
            if (location.Kvid == Guid.Empty)
            {
                (location.InternalCode.IsNullOrEmpty()).ThrowIfTrue("缺少目的地点配置InternalCode信息");
                location = conn.Single<Location>(o => o.InternalCode == location.InternalCode);
                (location == null).ThrowIfTrue($"未找到{location.InternalCode}目的地点配置信息!");
            }
            var exist = conn.SingleById<Sample>(sample.Kvid);
            exist.ThrowIfNull("未找到此样品信息!");
            (exist.PackageKvid == Guid.Empty).ThrowIfTrue("样品未打包,无需取出!");
            var package = conn.SingleById<Sample>(exist.PackageKvid);
            //var local = conn.Single<Location>(o => o.InternalCode == location.InternalCode);
            //(local == null).ThrowIfTrue($"未找到{location.InternalCode}目的地点配置信息!");
            var currentRoute = exist.GetCurrentRoute();//conn.Single<Route>(o => o.SampleKvid == existSample.Kvid && o.NextLocationKvid == Guid.Empty);

            IDbTransaction trans = null;//事务，如果外部来了Connection，不生成事务
            if (useTransaction)
            {
                trans = conn.OpenTransaction();
            }
            //var trans = conn.OpenTransaction();
            try
            {
                exist.PackageName = string.Empty;
                exist.AddOnlyProperties(o => o.PackageName);
                exist.PackageKvid = Guid.Empty;
                exist.AddOnlyProperties(o => o.PackageKvid);
                conn.UpdateOnly(exist);

                Route newRoute = new Route();
                if (route != null) newRoute.PopulateWith(route);
                newRoute.SampleRootKvid = exist.RootKvid;
                newRoute.SampleKvid = exist.Kvid;
                newRoute.SampleParentKvid = exist.ParentKvid;
                newRoute.SampleName = exist.Name;
                newRoute.BizId = exist.BizId;
                newRoute.BizKvid = exist.BizKvid;
                newRoute.BizType = exist.BizType;
                newRoute.Title = newRoute.Title.IsNullOrEmpty() ? $"样品[{exist.Name}]从[{(package.PackageName.IsNullOrEmpty() ? package.Name : $"{package.PackageName}({package.Name})")}]取出" : newRoute.Title;
                newRoute.Type = newRoute.Type.IsNullOrEmpty() ? "取样" : newRoute.Type;
                newRoute.CurrentLocationKvid = location.Kvid;
                newRoute.CurrentLocationTitle = location.Title;
                newRoute.OperateTime = newRoute.OperateTime == DateTime.MinValue ? DateTime.Now : newRoute.OperateTime;
                if (newRoute.OperatorName.IsNullOrEmpty())
                {
                    newRoute.OperatorName = KiviiContext.CurrentMember.FullName;
                    newRoute.OperatorKvid = KiviiContext.CurrentMember.Kvid;
                }
                conn.Insert(newRoute);

                if (currentRoute != null)
                {
                    currentRoute.NextLocationKvid = location.Kvid;
                    currentRoute.NextLocationTitle = location.Title;
                    currentRoute.AddOnlyProperties(o => o.NextLocationKvid);
                    currentRoute.AddOnlyProperties(o => o.NextLocationTitle);
                    conn.UpdateOnly(currentRoute);
                }

                trans?.Commit();
                if (package != null)
                {
                    if (!conn.Exists<Sample>(o => o.PackageKvid == package.Kvid && o.Type != SampleType.Package))
                    {
                        package.Status = (int)PackageStatus.Empty;
                        package.AddOnlyProperties(o => o.Status);
                        conn.UpdateOnly(package);
                    }
                }
            }
            catch (Exception ex)
            {
                trans?.Rollback();
                throw ex;
            }
            exist.RemoveAllOnlyProperties();
            return exist;
        }

        /// <summary>
        /// 样品入库
        /// </summary>
        /// <param name="sample"></param>
        /// <param name="location"></param>
        /// <param name="route"></param>
        /// <param name="conn"></param>
        /// <returns></returns>
        public static Sample Inbound(this Sample sample, Location location, Route route = null, IDbConnection conn = null)
        {
            (sample == null).ThrowIfTrue("请传入需要流转的样品信息!");
            (location == null).ThrowIfTrue("缺少目的地点配置信息!");
            bool useTransaction = conn == null;//是否启用内部事务
            if (conn == null) conn = KiviiContext.GetOpenedDbConnection<Sample>();
            //var conn = KiviiContext.GetOpenedDbConnection<Sample>();
            conn.InitEntityType<Location>();
            if (location.Kvid == Guid.Empty)
            {
                (location.InternalCode.IsNullOrEmpty()).ThrowIfTrue("缺少目的地点配置InternalCode信息");
                location = conn.Single<Location>(o => o.InternalCode == location.InternalCode);
                (location == null).ThrowIfTrue($"未找到{location.InternalCode}目的地点配置信息!");
            }
            var existSample = conn.SingleById<Sample>(sample.Kvid);
            existSample.ThrowIfNull("未找到此样品信息!");
            //找出当前样品所在路由地点,根据Next为空说明就在此地
            var currentRoute = existSample.GetCurrentRoute();//conn.Single<Route>(o => o.SampleKvid == existSample.Kvid && o.NextLocationKvid == Guid.Empty);

            IDbTransaction trans = null;//事务，如果外部来了Connection，不生成事务
            if (useTransaction)
            {
                trans = conn.OpenTransaction();
            }
            //var trans = conn.OpenTransaction();
            try
            {
                #region 生成新的路由信息,更新上一个路由信息
                if (route == null) route = new Route();
                route.SampleRootKvid = existSample.RootKvid;
                route.SampleKvid = existSample.Kvid;
                route.SampleParentKvid = existSample.ParentKvid;
                route.SampleName = existSample.Name;
                route.BizId = existSample.BizId;
                route.BizKvid = existSample.BizKvid;
                route.BizType = existSample.BizType;
                route.Title = route.Title.IsNullOrEmpty() ? $"样品[{existSample.Name}]入库.库位:{sample.PackageName}" : route.Title;
                route.Type = route.Type.IsNullOrEmpty() ? "入库" : route.Type;
                route.CurrentLocationKvid = location.Kvid;
                route.CurrentLocationTitle = location.Title;
                route.OperateTime = route.OperateTime == DateTime.MinValue ? DateTime.Now : route.OperateTime;
                existSample.AssignTime = route.OperateTime;
                if (route.OperatorName.IsNullOrEmpty())
                {
                    route.OperatorName = KiviiContext.CurrentMember.FullName;
                    route.OperatorKvid = KiviiContext.CurrentMember.Kvid;
                }
                conn.Insert(route);

                if (currentRoute != null)
                {
                    currentRoute.NextLocationKvid = location.Kvid;
                    currentRoute.NextLocationTitle = location.Title;
                    currentRoute.AddOnlyProperties(o => o.NextLocationKvid);
                    currentRoute.AddOnlyProperties(o => o.NextLocationTitle);
                    conn.UpdateOnly(currentRoute);
                }
                #endregion

                existSample.Status = (int)PackageStatus.Inbound;
                existSample.AddOnlyProperties(o => o.Status);
                if(!sample.PackageName.IsNullOrEmpty())
                {
                    existSample.PackageName = sample.PackageName;
                    existSample.AddOnlyProperties(o => o.PackageName);
                }
                conn.UpdateOnly(existSample);
                existSample.RemoveAllOnlyProperties();
                trans?.Commit();
            }
            catch (Exception ex)
            {
                trans?.Rollback();
                throw ex;
            }
            return existSample;
        }

        /// <summary>
        /// 获取当前指定样品所在地
        /// </summary>
        /// <param name="sample"></param>
        /// <returns></returns>
        public static Location GetCurrentLocation(this Sample sample)
        {
            if (sample == null) return null;
            //sample.ThrowIfNull("请传入要取出的样品信息!");
            var conn = KiviiContext.GetOpenedDbConnection<Sample>();
            var queryRoute = conn.From<Route>();
            queryRoute.Where(o => o.SampleKvid == sample.Kvid && o.NextLocationKvid == Guid.Empty);
            var routes = conn.Select(queryRoute);
            if (routes.IsNullOrEmpty()) return null;
            //routes.ThrowIfNullOrEmpty("未找到指定样品当前所在地点!");
            //(routes.Count() > 1).ThrowIfTrue("所在地点不唯一!");
            if (routes.Count() > 1) return null;
            var route = routes[0];
            sample.CurrentRoute = route;
            var rtns = conn.SingleById<Location>(route.CurrentLocationKvid);
            if (rtns == null) return null;
            sample.CurrentLocation = rtns;
            //rtns.ThrowIfNull($"未找到所在地({route.CurrentLocationTitle})数据信息!");
            return rtns;
        }
        /// <summary>
        /// 获取当前指定样品当前路由
        /// </summary>
        /// <param name="sample"></param>
        /// <returns></returns>
        public static Route GetCurrentRoute(this Sample sample)
        {
            if (sample == null) return null;
            //sample.ThrowIfNull("请传入要取出的样品信息!");
            var conn = KiviiContext.GetOpenedDbConnection<Sample>();
            var queryRoute = conn.From<Route>();
            queryRoute.Where(o => o.SampleKvid == sample.Kvid && o.NextLocationKvid == Guid.Empty);
            var routes = conn.Select(queryRoute);
            if (routes.IsNullOrEmpty()) return null;
            //routes.ThrowIfNullOrEmpty("未找到指定样品当前所在地点!");
            //(routes.Count() > 1).ThrowIfTrue("所在地点不唯一!");
            if (routes.Count() > 1) return null;
            var rtns = routes[0];
            sample.CurrentRoute = rtns;
            return rtns;
        }

        /// <summary>
        /// 获取当前指定样品的全部路由
        /// </summary>
        /// <param name="sample"></param>
        /// <returns></returns>
        public static List<Route> GetRoutes(this Sample sample)
        {
            if (sample == null) return null;
            //sample.ThrowIfNull("请传入要取出的样品信息!");
            var conn = KiviiContext.GetOpenedDbConnection<Sample>();
            var queryRoute = conn.From<Route>();
            queryRoute.Where(o => o.SampleKvid == sample.Kvid);
            queryRoute.OrderBy(o => o.CreateTime);
            var rtns = conn.Select(queryRoute);
            if (rtns.IsNullOrEmpty()) return null;
            sample.Routes = rtns;
            var currentRoute = rtns.FirstOrDefault(o => o.NextLocationKvid == Guid.Empty);
            if (currentRoute != null) sample.CurrentRoute = currentRoute;
            return rtns;
        }

        public static List<Sample> GetChildrens(this Sample sample)
        {
            if (sample == null) return null;
            var conn = KiviiContext.GetOpenedDbConnection<Sample>();
            var queryChildren = conn.From<Sample>();
            queryChildren.Where(o => o.ParentKvid == sample.Kvid);
            queryChildren.OrderBy(o => o.CreateTime);
            var rtns = conn.Select(queryChildren);
            if (rtns.IsNullOrEmpty()) return null;
            sample.Childrens = rtns;
            return rtns;
        }

        public static Sample GetCurrentPackage(this Sample sample)
        {
            if (sample == null) return null;
            if (sample.PackageKvid == Guid.Empty) return null;
            var conn = KiviiContext.GetOpenedDbConnection<Sample>();

            var rtns = conn.Single<Sample>(o => o.Kvid == sample.PackageKvid);
            if (rtns == null) return null;
            sample.CurrentPackage = rtns;
            return rtns;
        }

        public static Guid GetKvidByTid(this string tid)
        {
            //tid.ThrowIfNullOrEmpty("请传入Tid");
            if (tid.IsNullOrEmpty()) return Guid.Empty;
            var conn = KiviiContext.GetOpenedDbConnection<RfidBound>();
            var rfid = conn.Single<RfidBound>(o => o.Tid == tid);
            if (rfid == null) return Guid.Empty;
            //rfid.ThrowIfNull("未找到当前的Rfid信息!");
            if (rfid.Epc.IsNullOrEmpty()) return Guid.Empty;
            //rfid.Epc.ThrowIfNullOrEmpty("当前Rfid未绑定样品信息!");
            if (!Guid.TryParse(rfid.Epc, out var rtns)) return Guid.Empty;
            return rtns;
        }
    }
}
